/*
 * DESCRIPTION:
 * Common file for "game" directory.
*/

#pragma once

#include "../engine/circle.h"
#include "../engine/event.h"
#include "../engine/font.h"
#include "../engine/keyboard.h"
#include "../engine/mouse.h"
#include "../engine/music.h"
#include "../engine/rectangle.h"
#include "../engine/renderer.h"
#include "../engine/sound.h"
#include "../engine/text.h"
#include "../engine/tick.h"
#include "../engine/window.h"
#include "../util/list.h"
#include "../util/math.h"
#include "../util/random.h"

#define TEXTURE_COUNT 46
#define SOUND_COUNT 25
#define FONT_COUNT 6
#define MUSIC_COUNT 8

#define PLAYER_FATNESS_MAX 3
#define PLAYER_FATNESS_THRESHOLD 2

typedef enum TextureType
{
	TEXTURE_NONE,
	TEXTURE_LOGO,
	TEXTURE_GUI_CURSOR,
	TEXTURE_GUI_SNAKE,
	TEXTURE_GUI_REDIRECT,
	TEXTURE_GUI_CASH,
	TEXTURE_GUI_UPGRADE_SNAKE,
	TEXTURE_GUI_UPGRADE_LIFETIME,
	TEXTURE_GUI_UPGRADE_REDIRECT,
	TEXTURE_GUI_SIGN_SMALL,
	TEXTURE_GUI_SIGN_BIG,
	TEXTURE_SHADOW,
	TEXTURE_ARC,
	TEXTURE_PLAY_BG,
	TEXTURE_PLAYER_STAGE_ONE,
	TEXTURE_PLAYER_STAGE_TWO,
	TEXTURE_PLAYER_STAGE_THREE,
	TEXTURE_PLAYER_TAIL_STAGE_ONE,
	TEXTURE_PLAYER_TAIL_STAGE_TWO,
	TEXTURE_PLAYER_TAIL_STAGE_THREE,
	TEXTURE_PLAYER_SNAKE,
	TEXTURE_ENEMY_CUPPER,
	TEXTURE_ENEMY_CHIPPER,
	TEXTURE_ENEMY_CHIP,
	TEXTURE_ENEMY_CRUMBLER,
	TEXTURE_COLLECTIBLE_DOLLARS,
	TEXTURE_COLLECTIBLE_SACK,
	TEXTURE_COLLECTIBLE_BULLIONS,
	TEXTURE_COLLECTIBLE_MIRACLE_CURE,
	TEXTURE_COLLECTIBLE_DYNAMITE,
	TEXTURE_COLLECTIBLE_BERSERK,
	TEXTURE_INFO_THEME,
	TEXTURE_INFO_WILDCARD_ADMIN,
	TEXTURE_INFO_WILDCARD_PATREON,
	TEXTURE_INFO_WILDCARD_COMMUNITY,
	TEXTURE_INTRO_SCENE_ONE,
	TEXTURE_INTRO_SCENE_TWO,
	TEXTURE_INTRO_SCENE_THREE,
	TEXTURE_INTRO_SCENE_FOUR,
	TEXTURE_INTRO_SCENE_FIVE,
	TEXTURE_ENDING_FAIL,
	TEXTURE_ENDING_ONE,
	TEXTURE_ENDING_TWO,
	TEXTURE_ENDING_THREE,
	TEXTURE_TITLE_BACKGROUND,
	TEXTURE_RESULTS,
} TextureType;

typedef enum SoundType
{
	SOUND_NONE,
	SOUND_MENU,
	SOUND_BLIP,
	SOUND_SNAKE_FIRE,
	SOUND_SNAKE_REDIRECT,
	SOUND_SNAKE_RETURN,
	SOUND_ENEMY_HURT,
	SOUND_CUPPER_KILL,
	SOUND_CHIPPER_KILL,
	SOUND_CRUMBLER_KILL,
	SOUND_CHIP_KILL,
	SOUND_RUMBLE,
	SOUND_COLLECTIBLE,
	SOUND_DOLLARS,
	SOUND_SACK,
	SOUND_BULLIONS,
	SOUND_PURCHASE,
	SOUND_MIRACLE_CURE,
	SOUND_DYNAMITE,
	SOUND_BERSERK,
	SOUND_SWALLOW,
	SOUND_GROWTH,
	SOUND_ERROR,
	SOUND_GO,
	SOUND_SPIT,
} SoundType;

typedef enum MusicType
{
	MUSIC_NONE,
	MUSIC_WIND,
	MUSIC_PLAY,
	MUSIC_WAVE_CLEAR,
	MUSIC_WAVE_FAIL,
	MUSIC_RESULTS,
	MUSIC_TITLE,
	MUSIC_INTRO,
} MusicType;

typedef enum FontType
{
	FONT_NONE,
	FONT_TINY,
	FONT_SMALL,
	FONT_MEDIUM,
	FONT_BIG,
	FONT_MEGA,
} FontType;

typedef enum StateType
{
	STATE_NONE,
	STATE_TITLE,
	STATE_INTRO,
	STATE_PLAY,
	STATE_INFO,
	STATE_ENDING_FAIL,
	STATE_ENDING_ONE,
	STATE_ENDING_TWO,
	STATE_ENDING_THREE,
	STATE_RESULTS
} StateType;

typedef enum EnemyType
{
	ENEMY_NONE,
	ENEMY_CUPPER,
	ENEMY_CHIPPER,
	ENEMY_CRUMBLER,
	ENEMY_CHIP
} EnemyType;

typedef enum CollectibleType
{
	COLLECTIBLE_NONE,
	COLLECTIBLE_DOLLARS,
	COLLECTIBLE_SACK,
	COLLECTIBLE_BULLIONS,
	COLLECTIBLE_MIRACLE_CURE,
	COLLECTIBLE_DYNAMITE,
	COLLECTIBLE_BERSERK
} CollectibleType;

typedef enum PowerupType
{
	POWERUP_NONE,

} PowerupType;

typedef enum Direction
{
	DIRECTION_LEFT,
	DIRECTION_RIGHT,
	DIRECTION_UP,
	DIRECTION_DOWN
} Direction;

typedef enum SnakeFrameType
{
	SNAKE_NONE,
	SNAKE_REDIRECT,
	SNAKE_RETURN
} SnakeFrameType;

typedef enum ButtonStateType
{
	BUTTON_STATE_NONE,
	BUTTON_STATE_HOVER,
	BUTTON_STATE_HOVERING,
	BUTTON_STATE_PRESS,
	BUTTON_STATE_HELD
} ButtonStateType;

typedef enum EndingType
{
	ENDING_FAIL,
	ENDING_ONE,
	ENDING_TWO,
	ENDING_THREE
} EndingType;
	
typedef struct Game Game;

typedef struct Button
{
	Sprite sprite;
	ButtonStateType state;
} Button;

typedef struct Cursor
{
	Sprite sprite;
} Cursor;

typedef struct Snake
{
	Sprite sprite;
	Circle shadow;
	Sprite arc;
	f64 dX;
	f64 dY;
	s32 damage;
	s32 lifetime;
	s32 animationTimer;
	s32 returnTimer;
	s32 redirectCount;
	s32 redirectTimer;
	bool_u8 isReturning;
} Snake;

typedef struct Collectible
{
	Sprite sprite;
	Circle shadow;
	s32 lifetime;
	bool_u8 isActive;
	s32 activeTimer;
	f64 dX;
	f64 dY;
	CollectibleType type;
	bool_u8 isAttached;
	Snake* attachedSnake;
} Collectible;

typedef struct Enemy
{
	EnemyType type;
	Sprite sprite;
	Circle shadow;
	bool_u8 isActing;
	bool_u8 isActive;
	List snakesHit; /* Snake* */
	List returnSnakesHit; /* Snake* */
	f64 range;
	s32 attackTimer;
	s32 animationTimer;
	s32 lifetime;
	s32 hitTimer;
	s32 hitTimerMax;
	s32 health;
	f64 dX;
	f64 dY;
} Enemy;

typedef struct Message
{
	Sprite sprite;
	f64 x;
	f64 y;
	f64 dX;
	f64 dY;
	f64 dR;
	s32 lifetime;
} Message;

typedef struct Player
{
	Circle shadow;
	Direction direction;
	List snakes;
	Sprite sprite;
	Sprite tailSprite;
	bool_u8 isAbleToFire;
	bool_u8 isAbleToMove;
	bool_u8 isBeingFattened;
	bool_u8 isCashUpdate;
	bool_u8 isBeingDefeated;
	bool_u8 isDefeat;
	bool_u8 isFatnessUpdate;
	bool_u8 isLifetimeUpgradeUpdate;
	bool_u8 isPowerUp;
	bool_u8 isRedirectUpdate;
	bool_u8 isRedirectUpgradeUpdate;
	bool_u8 isSnakeUpdate;
	bool_u8 isSnakeUpgradeUpdate;
	bool_u8 isTailAnimate;
	bool_u8 isStationary;
	bool_u8 isKeyboardMove;
	f64 dX;
	f64 dY;
	s32 cash;
	s32 enemiesKilled;
	s32 fatness;
	s32 fatnessTimer;
	s32 cashCollected;
	s32 lifetime;
	s32 lifetimeUpgradeLevel;
	s32 powerUpTimer;
	s32 redirectMax;
	s32 redirectTimer;
	s32 redirectUpgradeLevel;
	s32 redirects;
	s32 snakeMax;
	s32 snakeReserve;
	s32 snakeUpgradeLevel;
	s32 defeatTimer;
	s32 tailTimer;
	s32 waveEnemiesKilled;
	s32 redirectRechargeTimer;
	s32 redirectUpgradePrice;
	s32 snakeUpgradePrice;
	s32 lifetimeUpgradePrice;
} Player;

typedef struct GUI
{
	Cursor cursor;
	Sprite snakeSprite;
	Sprite snakeText;
	Sprite redirectSprite;
	Sprite redirectText;
	Sprite cashSprite;
	Sprite cashText;
	Sprite snakeUpgradeText;
	Sprite lifetimeUpgradeText;
	Sprite redirectUpgradeText;
	Sprite upgradeText;
	Button snakeUpgradeButton;
	Button lifetimeUpgradeButton;
	Button redirectUpgradeButton;
	Button readyButton;
	Sprite waveText;
	Sprite snakeUpgradeBackgroundSprite;
	Sprite redirectUpgradeBackgroundSprite;
	Sprite lifetimeUpgradeBackgroundSprite;
	Sprite upgradeBackgroundSprite;
	Sprite readyBackgroundSprite;
	Sprite cashBackgroundSprite;
	Sprite cashTextBackgroundSprite;
	Sprite waveBackgroundSprite;
	Sprite bountyBackgroundSprite;
	Sprite bountyText;
} GUI;

typedef struct Play
{
	struct Game* g;
	Sprite pauseText;
	GUI gui;
	Player player;
	bool_u8 isWaveFailed;
	bool_u8 isWaveCleared;
	bool_u8 isGameCleared;
	bool_u8 isPaused;
	bool_u8 isBetweenWaves;
	bool_u8 isReady;
	bool_u8 isCountdown;
	bool_u8 isCountdownComplete;
	bool_u8 isWaveClearMusicComplete;
	bool_u8 isWaveUpdate;
	bool_u8 isUpgradesActive;
	bool_u8 isUpgradeUpdate;
	bool_u8 isBountyUpdate;
	bool_u8 isGameClearing;
	bool_u8 isGracePeriod;
	bool_u8 isTutorial;
	Sprite tutorialText;
	Sprite bg;
	s32 gameClearTimer;
	s32 countdownTimer;
	s32 tutorialTimer;
	s32 wave;
	s32 game;
	s32 spawnTimer;
	s32 enemyMax;
	s32 gracePeriodTimer;
	List enemies; /* Enemy */
	List spawnEnemies; /* EnemyType */
	List collectibles; /* Collectible */
	List messages; /* Message */
} Play;

typedef struct Title
{
	struct Game* g;
	Cursor cursor;
	Sprite startBackgroundSprite;
	Sprite infoBackgroundSprite;
	Sprite backgroundSprite;
	Button startButton;
	Button infoButton;
	bool_u8 isInfoPressed;
	bool_u8 isStartPressed;
	Sprite logoSprite;
} Title;

typedef struct Scene
{
	char* text;
	TextureType textureType;
} Scene;

typedef struct SceneHandler
{
	struct Game* g;
	Scene* scenes;
	Sprite sprite;
	Sprite text;
	bool_u8 isTextEnd;
	bool_u8 isEnd;
	s32 updateTimer;
	s32 sceneCount;
	s32 sceneIndex;
	s32 textIndex;
	s32 textLength;
} SceneHandler;

typedef struct Intro
{
	Game* g;
	SceneHandler sceneHandler;
} Intro;

typedef struct Info
{
	struct Game* g;
	bool_u8 isBackPressed;
	Sprite licenseText;
	Sprite themeText;
	Sprite themeSprite;
	Sprite wildcardAdminSprite;
	Sprite wildcardAdminText;
	Sprite wildcardPatreonSprite;
	Sprite wildcardPatreonText;
	Sprite wildcardCommunitySprite;
	Sprite wildcardCommunityText;
	Sprite backBackgroundSprite;
	Button backButton;
	Cursor cursor;
} Info;

typedef struct Ending
{
	Game* g;
	SceneHandler sceneHandler;
	EndingType type;
	Button plusButton;
	Button endButton;
	bool_u8 isActualEnding;
	bool_u8 isNewGamePlus;
	bool_u8 isResults;
	Cursor cursor;
} Ending;

typedef struct Results
{
	struct Game* g;
	bool_u8 isEnd;
	Sprite highestWave;
	Sprite cashCollected;
	Sprite cashHeld;
	Sprite bountiesHunted;
	Sprite thanksText;
	Sprite resultsText;
	Sprite sprite;
} Results;

typedef struct Game
{
	Ending ending;
	Event event;
	Info info;
	Intro intro;
	Keyboard keyboard;
	Mix_Chunk* sounds[SOUND_COUNT];
	Mix_Music* music[MUSIC_COUNT];
	Mouse mouse;
	Play play;
	Results results;
	SDL_Renderer* renderer;
	SDL_Texture* textures[TEXTURE_COUNT];
	SDL_Window* window;
	StateType state;
	StateType stateNext;
	TTF_Font* fonts[FONT_COUNT];
	Tick tick;
	Title title;
	bool_u8 isStateTransition;
	bool_u8 isStateTransitioning;
	bool_u8 isNextStateWillBeInit;
	bool_u8 isPrevStateWillBeFreed;
	bool_u8 isNextStateResourcesWillBeInit;
	bool_u8 isPrevStateResourcesWillBeFreed;
	bool_u8 isFullScreen;
	s32 stateTimer;
} Game;
