/*
 * DESCRIPTION:
 * Game states.
*/

#include "game_state.h"

static void game_state_transition_update(Game* _g);
static void game_state_set(Game* _g, StateType _s);

/* Update state change transition. */
static void
game_state_transition_update(Game* _g)
{
	if (_g->isStateTransitioning)
	{
		_g->stateTimer++;

		if (_g->stateTimer >= STATE_TRANSITION_TIMER_MAX)
		{
			game_state_set(_g, _g->stateNext);
			_g->stateNext = STATE_NONE;
			_g->isStateTransitioning = FALSE;
		}
	}
	else
	{
		_g->stateTimer--;

		if (_g->stateTimer <= 0)
			_g->isStateTransition = FALSE;
	}
}

/* Sets the game's state to a specified one. */
static void
game_state_set(Game* _g, StateType _s)
{
	game_state_free(_g);

	_g->state = _s;

	if (_g->isNextStateResourcesWillBeInit)
	{
		switch (_g->state)
		{
			case STATE_TITLE:
				game_title_init(_g);
				break;
			case STATE_INTRO:
				game_intro_init(_g);
				break;
			case STATE_PLAY:
				game_play_init(_g);
				break;
			case STATE_INFO:
				game_info_init(_g);
				break;
			case STATE_ENDING_FAIL:
			case STATE_ENDING_ONE:
			case STATE_ENDING_TWO:
			case STATE_ENDING_THREE:
				game_ending_init(_g);
				break;
			case STATE_RESULTS:
				game_results_init(_g);
				break;
			default:
				break;
		}

		_g->isNextStateResourcesWillBeInit = FALSE;
	}

	if (_g->isNextStateWillBeInit)
	{
		switch (_g->state)
		{
			case STATE_TITLE:
				title_init(&_g->title, _g);
				break;
			case STATE_INTRO:
				intro_init(&_g->intro, _g);
				break;
			case STATE_PLAY:
 			       	play_init(&_g->play, _g);
				break;
			case STATE_INFO:
 			       	info_init(&_g->info, _g);
				break;
			case STATE_ENDING_FAIL:
				ending_init(&_g->ending, _g, ENDING_FAIL);
				break;
			case STATE_ENDING_ONE:
				ending_init(&_g->ending, _g, ENDING_ONE);
				break;
			case STATE_ENDING_TWO:
				ending_init(&_g->ending, _g, ENDING_TWO);
				break;
			case STATE_ENDING_THREE:
				ending_init(&_g->ending, _g, ENDING_THREE);
				break;
			case STATE_RESULTS:
				results_init(&_g->results, _g);
				break;
			default:
				break;
		}

		_g->isNextStateWillBeInit = FALSE;
	}
}

/* Transition the a state type with a fade-in/fade-out. */
void
game_state_transition(Game* _g, StateType _s, bool_u8 _i, bool_u8 _f, bool_u8 _rI, bool_u8 _rF)
{
	_g->stateTimer = 0;
	_g->stateNext = _s;
	_g->isNextStateWillBeInit = _i;
	_g->isPrevStateWillBeFreed = _f;
	_g->isNextStateResourcesWillBeInit = _rI;
	_g->isPrevStateResourcesWillBeFreed = _rF;
	_g->isStateTransitioning = TRUE;
	_g->isStateTransition = TRUE;
}

/* Frees everything related to the game's current state. */
void
game_state_free(Game* _g)
{
	if (_g->isPrevStateResourcesWillBeFreed)
	{
		switch (_g->state)
		{
			case STATE_TITLE:
				game_title_free(_g);
				break;
			case STATE_INTRO:
				game_intro_free(_g);
				break;
			case STATE_PLAY:
				game_play_free(_g);
				break;
			case STATE_INFO:
				game_info_free(_g);
				break;
			case STATE_ENDING_FAIL:
			case STATE_ENDING_ONE:
			case STATE_ENDING_TWO:
			case STATE_ENDING_THREE:
				game_ending_free(_g);
				break;
			case STATE_RESULTS:
				game_results_free(_g);
				break;
			default:
				break;

		}

		_g->isPrevStateResourcesWillBeFreed = FALSE;
	}

	if (_g->isPrevStateWillBeFreed)
	{
		switch (_g->state)
		{
			case STATE_TITLE:
				title_free(&_g->title);
				break;
			case STATE_INTRO:
				intro_free(&_g->intro);
				break;
			case STATE_PLAY:
				play_free(&_g->play);
				break;
			case STATE_INFO:
				info_free(&_g->info);
				break;
			case STATE_ENDING_FAIL:
			case STATE_ENDING_ONE:
			case STATE_ENDING_TWO:
			case STATE_ENDING_THREE:
				ending_free(&_g->ending);
				break;
			case STATE_RESULTS:
				results_free(&_g->results);
				break;
			default:
				break;
		}

		_g->isPrevStateWillBeFreed = FALSE;
	}
}

/* Updates the game's state. */
void
game_state_update(Game* _g)
{
	if (_g->isStateTransition)
		game_state_transition_update(_g);

	switch (_g->state)
	{
		case STATE_TITLE:
			game_title_update(_g);

			if 
			(
				_g->title.isStartPressed &&
				!_g->isStateTransition
			)
			{
				_g->title.isStartPressed = FALSE;
				game_state_transition(_g, STATE_INTRO, TRUE, TRUE, TRUE, TRUE);
			}

			if 
			(
				_g->title.isInfoPressed &&
				!_g->isStateTransition
			)
			{
				_g->title.isInfoPressed = FALSE;
				game_state_transition(_g, STATE_INFO, TRUE, TRUE, TRUE, TRUE);
			}
			break;
		case STATE_INTRO:
			game_intro_update(_g);

			if
			(
				_g->intro.sceneHandler.isEnd &&
				!_g->isStateTransition
			)
				game_state_transition(_g, STATE_PLAY, TRUE, TRUE, TRUE, TRUE);
			break;
		case STATE_PLAY:
			game_play_update(_g);


			if
			(
				keyboard_press(&_g->keyboard, KEYBOARD_ESCAPE) &&
				_g->play.isPaused &&
				!_g->isStateTransition
			)
				game_state_transition(_g, STATE_TITLE, TRUE, TRUE, TRUE, TRUE);
			else if 
			(
				_g->play.player.isDefeat &&
				!_g->isStateTransition
			)
				game_state_transition(_g, STATE_ENDING_FAIL, TRUE, TRUE, TRUE, TRUE);
			else if
			(
				_g->play.isGameCleared &&
				!_g->isStateTransition
			)
			{
				if (_g->play.player.cash >= ENDING_THREE_CASH_THRESHOLD)
					game_state_transition(_g, STATE_ENDING_THREE, TRUE, FALSE, TRUE, FALSE);
				else if (_g->play.player.cash >= ENDING_TWO_CASH_THRESHOLD)
					game_state_transition(_g, STATE_ENDING_TWO, TRUE, FALSE, TRUE, FALSE);
				else
					game_state_transition(_g, STATE_ENDING_ONE, TRUE, FALSE, TRUE, FALSE);
			}
			break;
		case STATE_INFO:
			game_info_update(_g);

			if
			(
				_g->info.isBackPressed &&
			 	!_g->isStateTransition
			)
			{
				_g->info.isBackPressed = FALSE;
				game_state_transition(_g, STATE_TITLE, TRUE, TRUE, TRUE, TRUE);
			}
			break;
		case STATE_ENDING_FAIL:
			game_ending_update(_g);

			if
			(
				_g->ending.sceneHandler.isEnd &&
				!_g->isStateTransition
			)
				game_state_transition(_g, STATE_RESULTS, TRUE, TRUE, TRUE, TRUE);
			break;
		case STATE_ENDING_ONE:
		case STATE_ENDING_TWO:
		case STATE_ENDING_THREE:
			game_ending_update(_g);

			if (!_g->isStateTransition)
			{
				if (_g->ending.isNewGamePlus)
				{
					game_state_transition(_g, STATE_PLAY, FALSE, TRUE, FALSE, FALSE);
					wave_new_game_plus(&_g->play);
					_g->ending.isNewGamePlus = FALSE;
					_g->ending.isActualEnding = FALSE;
				}
				else if (_g->ending.isResults)
			{
					game_state_transition(_g, STATE_RESULTS, TRUE, TRUE, TRUE, TRUE);
					_g->ending.isResults = FALSE;
					_g->ending.isActualEnding = FALSE;
				}
			}

			break;
		case STATE_RESULTS:
			game_results_update(_g);

			if
			(
				_g->results.isEnd &&
				!_g->isStateTransition
			)
				game_state_transition(_g, STATE_TITLE, TRUE, TRUE, TRUE, TRUE);
			break;
		default:
			break;
	}
}

/* Draws the game's state. */
void
game_state_draw(Game* _g)
{
	switch (_g->state)
	{
		case STATE_TITLE:
			game_title_draw(_g);
			break;
		case STATE_INTRO:
			game_intro_draw(_g);
			break;
		case STATE_PLAY:
			game_play_draw(_g);
			break;
		case STATE_INFO:
			game_info_draw(_g);
			break;
		case STATE_ENDING_FAIL:
		case STATE_ENDING_ONE:
		case STATE_ENDING_TWO:
		case STATE_ENDING_THREE:
			game_ending_draw(_g);
			break;
		case STATE_RESULTS:
			game_results_draw(_g);
			break;
		default:
			break;
	}

	if (_g->isStateTransition)
	{
		f64 progress;
		SDL_Rect rect;
		SDL_Color color;

		progress = (f64)_g->stateTimer / STATE_TRANSITION_TIMER_MAX;

		rect.x = 0;
		rect.y = 0;
		rect.w = STATE_TRANSITION_RECTANGLE_WIDTH;
		rect.h = STATE_TRANSITION_RECTANGLE_HEIGHT;

		color = STATE_TRANSITION_COLOR;
		color.a = SDL_ALPHA_OPAQUE * progress;

		rectangle_draw(&rect, _g->renderer, color);
	}
}
