/*
 * DESCRIPTION:
 * AI for the "Chipper" enemy.
*/

#include "chipper.h"

static s32 chipper_health_get(Enemy* _e, List* _l, Play* _p);
static s32 chipper_attack_timer_max_get(Enemy* _e, List* _l, Play* _p);

/* Gets the Chipper health. */
static s32
chipper_health_get(Enemy* _e, List* _l, Play* _p)
{
	s32 health;

	health = (s32)(CHIPPER_HEALTH_BASE + ((_p->game - 1) * CHIPPER_HEALTH_GAME_MULTIPLIER));

	health = CEIL(health, CHIPPER_HEALTH_MAX);

	return health;
}

/* Gets the Chipper attack timer max. */
static s32
chipper_attack_timer_max_get(Enemy* _e, List* _l, Play* _p)
{
	s32 attackTimerMax;

	attackTimerMax = (s32)(CHIPPER_ATTACK_TIMER_MAX_BASE - ((_p->game - 1) * CHIPPER_ATTACK_TIMER_MAX_GAME_MULTIPLIER));

	attackTimerMax = FLOOR(attackTimerMax, CHIPPER_ATTACK_TIMER_MAX_MIN);

	return attackTimerMax;
}

/* Initializes Chipper. */
void
chipper_init(Enemy* _e, List* _l, Play* _p)
{
	Atlas atlas;

	atlas_init
	(
		&atlas,
		_p->g->textures[TEXTURE_ENEMY_CHIPPER],
		CHIPPER_ROWS,
		CHIPPER_COLUMNS,
		CHIPPER_WIDTH,
		CHIPPER_HEIGHT
	);

	sprite_atlas_init
	(
		&_e->sprite,
		&atlas,
		ORIGIN_MIDDLE,
		CHIPPER_WIDTH,
		CHIPPER_HEIGHT
	);

	_e->shadow.r = CHIPPER_SHADOW_RADIUS;

	_e->sprite.frame = 2;
}

/* Sets the Chipper. */
void
chipper_set(Enemy* _e, List* _l, f64 _x, f64 _y, Play* _p)
{
	_e->sprite.x = _x;
	_e->sprite.y = _y;

	_e->shadow.x = _e->sprite.x + CHIPPER_SHADOW_OFFSET_X;
	_e->shadow.y = _e->sprite.y + CHIPPER_SHADOW_OFFSET_Y;

	_e->isActive = TRUE;
	_e->isActing = FALSE;
	_e->range = random_f64_in_range_get(CHIPPER_RANGE_MIN, CHIPPER_RANGE_MAX);
	_e->attackTimer = chipper_attack_timer_max_get(_e, _l, _p);
	_e->health = chipper_health_get(_e, _l, _p);
}

/* 
 * Updates Chipper.
 * Basically, it's going to slowly approach the player,
 * get into range of them,
 * and fire projectiles (chips) at the player.
*/
void
chipper_update(Enemy* _e, List* _l, Play* _p)
{
	f64 distance;
	f64 angle;
	SDL_Rect rect;
	SDL_Rect windowRect;

	rect = sprite_rect_get(&_e->sprite);
	windowRect = window_rect_get(_p->g->window);

	distance = math_distance
	(
		_e->shadow.x,
		_e->shadow.y,
		_p->player.shadow.x,
		_p->player.shadow.y
	);

	angle = math_atan
	(
		_e->shadow.x,
		_e->shadow.y,
		_p->player.shadow.x,
		_p->player.shadow.y		
	);

	if (_p->player.sprite.x < _e->sprite.x)
		_e->sprite.flip = FLIP_HORIZONTAL;
	else
		_e->sprite.flip = FLIP_NONE;

	/* Is this Chipper within range, and is fully on screen? */
	if 
	(
		distance <= _e->range &&
		SDL_HasIntersection(&rect, &windowRect)
	)
	{
		if (!_e->isActing)
			_e->animationTimer = 0;
		_e->isActing = TRUE;
	}
	else
	{
		if (_e->isActing)
			_e->animationTimer = 0;
		_e->isActing = FALSE;
	}

	_e->animationTimer--;
	
	if (_e->isActing) /* Waiting to shoot */
	{
		_e->dX = 0;
		_e->dY = 0;

		_e->sprite.frame = 1;

		_e->attackTimer--;

		if (_e->animationTimer > 0)
		{
			f64 progress;

			_e->sprite.frame = 3;

			progress = 1 - ((f64)_e->animationTimer / CHIPPER_ANIMATION_TIMER_MAX);

			_e->sprite.squash = SPRITE_SQUASH_DEFAULT + (cos(ROTATION_MAX_RAD * progress) * CHIPPER_SQUASH_MULTIPLIER);
			_e->sprite.stretch = SPRITE_STRETCH_DEFAULT + (cos(ROTATION_MAX_RAD * progress) * CHIPPER_STRETCH_MULTIPLIER);

			if (_e->animationTimer <= 0)
			{
				_e->sprite.squash = SPRITE_SQUASH_DEFAULT;
				_e->sprite.stretch = SPRITE_STRETCH_DEFAULT;
			}
		}

		if (_e->attackTimer <= 0) /* Just started shooting */
		{
			Enemy* e;

			e = enemy_add(_l, ENEMY_CHIP, _p);
			enemy_set(_l, e, _e->shadow.x, _e->shadow.y, _p);

			_e->attackTimer = chipper_attack_timer_max_get(_e, _l, _p);

			_e->animationTimer = CHIPPER_ANIMATION_TIMER_MAX;

			_e->sprite.frame = 3;

			_p->enemyMax++;

			sound_play(_p->g->sounds[SOUND_SPIT], -1);
		}
	}
	else /* Walk */
	{
		_e->dX = -cos(angle) * CHIPPER_VELOCITY;
		_e->dY = -sin(angle) * CHIPPER_VELOCITY;

		if (_e->animationTimer % CHIPPER_ANIMATION_MODULO == 0)
		{	
			_e->sprite.frame++;

			if (_e->sprite.frame > 2)
				_e->sprite.frame = 0;
		}

		if (_e->animationTimer <= 0)
			_e->animationTimer = CHIPPER_ANIMATION_TIMER_MAX;
	}

	_e->sprite.x += _e->dX;
	_e->sprite.y += _e->dY;

	_e->shadow.x = _e->sprite.x + CHIPPER_SHADOW_OFFSET_X;
	_e->shadow.y = _e->sprite.y + CHIPPER_SHADOW_OFFSET_Y;
}
