/*
 * DESCRIPTION:
 * AI for the "Cupper" enemy.
*/

#include "cupper.h"

static s32 cupper_health_get(Enemy* _e, List* _l, Play* _p);
static s32 cupper_attack_timer_max_get(Enemy* _e, List* _l, Play* _p);

/* Returns health of a Cupper. */
static s32
cupper_health_get(Enemy* _e, List* _l, Play* _p)
{
	s32 health;

	health = CUPPER_HEALTH_BASE + ((_p->game - 1) * CUPPER_HEALTH_GAME_MULTIPLIER);

	health = CEIL(health, CUPPER_HEALTH_MAX);

	return health;
}

/* Returns the attack timer max (period in which an attack is performed) of a Cupper. */
static s32
cupper_attack_timer_max_get(Enemy* _e, List* _l, Play* _p)
{
	s32 attackTimerMax;

	attackTimerMax = CUPPER_ATTACK_TIMER_MAX_BASE - (_p->game * CUPPER_ATTACK_TIMER_MAX_GAME_MULTIPLIER);

	attackTimerMax = FLOOR(attackTimerMax, CUPPER_ATTACK_TIMER_MAX_MIN);

	return attackTimerMax;
}

/* Initializes the Cupper. */
void
cupper_init(Enemy* _e, List* _l, Play* _p)
{
	Atlas atlas;

	atlas_init
	(
	 	&atlas,
		_p->g->textures[TEXTURE_ENEMY_CUPPER],
		CUPPER_ROWS,
		CUPPER_COLUMNS,
		CUPPER_WIDTH,
		CUPPER_HEIGHT
	);

	sprite_atlas_init
	(
		&_e->sprite,
		&atlas,
		ORIGIN_MIDDLE,
		CUPPER_WIDTH,
		CUPPER_HEIGHT
	);

	_e->shadow.r = CUPPER_SHADOW_RADIUS;
}

/* Sets the Cupper. */
void
cupper_set(Enemy* _e, List* _l, f64 _x, f64 _y, Play* _p)
{
	_e->sprite.x = _x;
	_e->sprite.y = _y;

	_e->shadow.x = _e->sprite.x + CUPPER_SHADOW_OFFSET_X;
	_e->shadow.y = _e->sprite.y + CUPPER_SHADOW_OFFSET_Y;

	_e->isActive = TRUE;
	
	_e->health = cupper_health_get(_e, _l, _p);
	_e->attackTimer = cupper_attack_timer_max_get(_e, _l, _p);
}

/* 
 * Updates Cupper.
 * Slowly hops towards the player.
*/
void
cupper_update(Enemy* _e, List* _l, Play* _p)
{
	f64 angle;
	f64 progress;

	_e->attackTimer--;
	_e->animationTimer--;

	progress = 1 - ((f64)_e->animationTimer / CUPPER_ANIMATION_TIMER_MAX);

	if (_e->animationTimer <= 0)
		_e->animationTimer = CUPPER_ANIMATION_TIMER_MAX;

	if (_e->isActing)
	{
		if (_e->attackTimer <= 0)
		{
			_e->dX = 0;
			_e->dY = 0;
			_e->isActing = FALSE;
			_e->attackTimer = cupper_attack_timer_max_get(_e, _l, _p);
		}
	}
	else
	{
		if (_e->attackTimer <= 0)
		{
			f64 angle;

			angle = math_atan
			(
				_e->shadow.x,
				_e->shadow.y,
				_p->player.shadow.x,
				_p->player.shadow.y		
			);

			_e->dX = -cos(angle) * CUPPER_VELOCITY;
			_e->dY = -sin(angle) * CUPPER_VELOCITY;

			_e->isActing = TRUE;

			_e->attackTimer = CUPPER_JUMP_TIMER_MAX_BASE;

			if (_p->player.sprite.x < _e->sprite.x)
				_e->sprite.flip = FLIP_HORIZONTAL;
			else
				_e->sprite.flip = FLIP_NONE;

			_e->sprite.frame = 0;
		}
		else if (_e->attackTimer <= CUPPER_ATTACK_TIMER_THRESHOLD)
			_e->sprite.frame = 1;
		else
		{
			_e->sprite.squash = (SPRITE_SQUASH_DEFAULT + cos((ROTATION_MAX_RAD * progress)) * CUPPER_SQUASH_MULTIPLIER);
			_e->sprite.stretch = (SPRITE_STRETCH_DEFAULT + cos((ROTATION_MAX_RAD * progress)) * CUPPER_STRETCH_MULTIPLIER);
		}
	}

	_e->sprite.x += _e->dX;
	_e->sprite.y += _e->dY;

	_e->shadow.x = _e->sprite.x + CUPPER_SHADOW_OFFSET_X;
	_e->shadow.y = _e->sprite.y + CUPPER_SHADOW_OFFSET_Y;
}
