/*
 * DESCRIPTION:
 * Game enemies.
*/

#include "enemy.h"

static CollectibleType enemy_collectible_type_get(Enemy* _e, List* _l, Play* _p);
static f64 enemy_collectible_drop_chance_get(Enemy* _e, List* _l, Play* _p);
static void enemy_collectible_drop(Enemy* _e, List* _l, Play* _p);

/* The enemy has dropped a collectible; what type is it? */
static CollectibleType
enemy_collectible_type_get(Enemy* _e, List* _l, Play* _p)
{
	f64 random;
	CollectibleType type;

	random = random_f64_get();

	if (random - ENEMY_COLLECTIBLE_POWERUP_DROP_CHANCE < 0)
		type = (CollectibleType)random_s32_in_range_get(COLLECTIBLE_MIRACLE_CURE, COLLECTIBLE_BERSERK);
	else if (random - ENEMY_COLLECTIBLE_BULLIONS_DROP_CHANCE < 0)
		type = COLLECTIBLE_BULLIONS;
	else if (random - ENEMY_COLLECTIBLE_SACK_DROP_CHANCE < 0)
		type = COLLECTIBLE_SACK;
	else
		type = COLLECTIBLE_DOLLARS;

	return type;
}

/* % of enemy dropping a collectible */
static f64
enemy_collectible_drop_chance_get(Enemy* _e, List* _l, Play* _p)
{
	f64 chance;

	chance = ENEMY_COLLECTIBLE_DROP_CHANCE_BASE + ((_p->game - 1) * ENEMY_COLLECTIBLE_DROP_CHANCE_GAME_MULTIPLIER);

	switch (_e->type)
	{
		case ENEMY_CHIPPER:
			chance += ENEMY_CHIPPER_DROP_CHANCE_BONUS;
			break;
		case ENEMY_CRUMBLER:
			chance += ENEMY_CRUMBLER_DROP_CHANCE_BONUS;
			break;
		case ENEMY_NONE:
		case ENEMY_CUPPER:
		default:
			break;
	}

	chance = CEIL(chance, ENEMY_COLLECTIBLE_DROP_CHANCE_MAX);

	return chance;
}

/* Will the enemy drop a collectible? */
static void
enemy_collectible_drop(Enemy* _e, List* _l, Play* _p)
{
	f64 chanceMax;
	f64 chance;
	
	chanceMax = enemy_collectible_drop_chance_get(_e, _l, _p);

	chance = random_f64_get();

	if (chance <= chanceMax)
	{
		CollectibleType type;

		type = enemy_collectible_type_get(_e, _l, _p);
		
		collectible_add(&_p->collectibles, type, _e->shadow.x, _e->shadow.y, _p);
	}
}

/* Adds an enemy to the game. */
Enemy*
enemy_add(List* _l, EnemyType _t, Play* _p)
{
	Enemy enemy;

	memset(&enemy, '\0', sizeof(Enemy));

	enemy.type = _t;

	list_init(&enemy.snakesHit);
	list_init(&enemy.returnSnakesHit);

	switch (enemy.type)
	{
		case ENEMY_CUPPER:
			cupper_init(&enemy, _l, _p);
			break;
		case ENEMY_CHIPPER:
			chipper_init(&enemy, _l, _p);
			break;
		case ENEMY_CHIP:
			chip_init(&enemy, _l, _p);
			break;
		case ENEMY_CRUMBLER:
			crumbler_init(&enemy, _l, _p);
			break;
		case ENEMY_NONE:
		default:
			break;
	}

	_p->isBountyUpdate = TRUE;

	return (Enemy*)list_append_new(_l, (void*)&enemy, sizeof(Enemy));
}

/* Removes an enemy from the game. */
void
enemy_remove(List* _l, Enemy* _e, Play* _p)
{
	list_free(&_e->snakesHit);
	list_free(&_e->returnSnakesHit);

	_p->isBountyUpdate = TRUE;
	_p->enemyMax--;

	list_node_from_data_remove(_l, _e, sizeof(Enemy));
}

/* Positions the enemy and prepares it for battle. */
void
enemy_set(List* _l, Enemy* _e, f64 _x, f64 _y, Play* _p)
{
	switch (_e->type)
	{
		case ENEMY_CUPPER:
			cupper_set(_e, _l, _x, _y, _p);
			break;
		case ENEMY_CHIPPER:
			chipper_set(_e, _l, _x, _y, _p);
			break;
		case ENEMY_CHIP:
			chip_set(_e, _l, _x, _y, _p);
			break;
		case ENEMY_CRUMBLER:
			crumbler_set(_e, _l, _x, _y, _p);
			break;
		case ENEMY_NONE:
		default:
			break;
	}
}

/* The enemy takes the specified amount of damage. */
void
enemy_damage(Enemy* _e, s32 _d, List* _l, Play* _p)
{
	_e->health -= _d;
	_e->hitTimerMax += _d * ENEMY_HIT_TIMER_MAX_MULTIPLIER;
	_e->hitTimer += _d * ENEMY_HIT_TIMER_MAX_MULTIPLIER;

	message_damage_add(_e->sprite.x, _e->sprite.y, _d, _p);
	sound_play(_p->g->sounds[SOUND_ENEMY_HURT], 0);

	if (_e->health <= 0)
	{
		_e->health = 0;
		enemy_defeat(_e, _l, _p);
	}
}

/* The enemy is defeated.. */
void
enemy_defeat(Enemy* _e, List* _l, Play* _p)
{
	if (_e->type != ENEMY_CHIP)
		enemy_collectible_drop(_e, _l, _p);

	switch (_e->type)
	{
		case ENEMY_CUPPER:
			sound_play(_p->g->sounds[SOUND_CUPPER_KILL], -1);
			break;
		case ENEMY_CHIPPER:
			sound_play(_p->g->sounds[SOUND_CHIPPER_KILL], -1);
			break;
		case ENEMY_CRUMBLER:
			sound_play(_p->g->sounds[SOUND_CRUMBLER_KILL], -1);
			break;
		case ENEMY_CHIP:
			sound_play(_p->g->sounds[SOUND_CHIP_KILL], -1);
			break;
		default:
			break;
	}

	_p->player.waveEnemiesKilled++;
	_p->player.enemiesKilled++;

	enemy_remove(_l, _e, _p);
}

/* Updates enemy. */
void
enemy_update(Enemy* _e, List* _l, Play* _p)
{
	switch (_e->type)
	{
		case ENEMY_CUPPER:
			cupper_update(_e, _l, _p);
			break;
		case ENEMY_CHIPPER:
			chipper_update(_e, _l, _p);
			break;
		case ENEMY_CHIP:
			chip_update(_e, _l, _p);
			break;
		case ENEMY_CRUMBLER:
			crumbler_update(_e, _l, _p);
			break;
		default:
			break;
	}

	if (_e->hitTimer > 0)
	{
		f64 progress;

		progress = 1 - ((f64)_e->hitTimer / _e->hitTimerMax);

		_e->sprite.color.r = SDL_ALPHA_OPAQUE * progress;
		_e->sprite.color.g = 0;
		_e->sprite.color.b = 0;

		_e->hitTimer--;

		if (_e->hitTimer <= 0)
		{
			_e->hitTimer = 0;
			_e->hitTimerMax = 0;
			_e->sprite.color = SPRITE_COLOR_DEFAULT;
		}
	}
}
