/*
 * DESCRIPTION:
 * Player snake code. 
*/

#include "player_snakes.h"

static Sprite snakeSprite;
static Sprite arcSprite;

static void player_snake_remove(Snake* _s, Player* _p, Play* _l);
static void player_snake_update(Snake* _s, Player* _p, Play* _l);
static void player_snake_draw(Snake* _s, Player* _p, Play* _l);

/* Removes a snake. */
static void
player_snake_remove(Snake* _s, Player* _p, Play* _l)
{
	Enemy** enemies;
        Collectible** collectibles;
        s32 enemyCount;
        s32 collectibleCount;

        enemies = (Enemy**)list_data_get(&_l->enemies);
        enemyCount = _l->enemies.count;

        /*
         * If you're looking for it, this is the most inefficent code in this entire thing.
         * And that's saying a lot!
        */
        for (s32 i = 0; i < enemyCount; i++)
        {
                Enemy* e;

                e = enemies[i];

                list_node_from_data_remove(&e->snakesHit, &_s, sizeof(Snake*));
                list_node_from_data_remove(&e->returnSnakesHit, &_s, sizeof(Snake*));
        }

        free(enemies);

        collectibles = (Collectible**)list_data_get(&_l->collectibles);
        collectibleCount = _l->collectibles.count;

        for (s32 i = 0; i < collectibleCount; i++)
        {
                Collectible* c;

                c = collectibles[i];

                if (c->attachedSnake == _s)
                {
                        c->attachedSnake = NULL;
                        c->isAttached = FALSE;
			collectible_collect(c, &_l->collectibles, _l);
                }
        }

	free(collectibles);

        list_node_from_data_remove(&_p->snakes, _s, sizeof(Snake));

        _p->snakeReserve++;

        _p->snakeReserve = CEIL(_p->snakeReserve, _p->snakeMax);

        _p->isSnakeUpdate = TRUE;

	if (_p->redirects < _p->redirectMax)
	{
		_p->redirects++;
		_p->isRedirectUpdate = TRUE;
	}
}

/* Updates a given snake. */
static void
player_snake_update(Snake* _s, Player* _p, Play* _l)
{
	Collectible** collectibles;
	Enemy** enemies;
	s32 enemyCount;
	s32 collectibleCount;

	_s->sprite.x += _s->dX;
	_s->sprite.y += _s->dY;

	_s->shadow.x = _s->sprite.x + PLAYER_SNAKE_SHADOW_OFFSET_X;
	_s->shadow.y = _s->sprite.y + PLAYER_SNAKE_SHADOW_OFFSET_Y;

	_s->animationTimer--;

	enemies = (Enemy**)list_data_get(&_l->enemies);
	enemyCount = _l->enemies.count;

	if (_s->animationTimer <= 0)
	{
		if (_s->sprite.flip == FLIP_VERTICAL)
			_s->sprite.flip = FLIP_NONE;
		else
			_s->sprite.flip = FLIP_VERTICAL;

		_s->animationTimer = PLAYER_SNAKE_ANIMATION_TIMER_MAX;
	}

	for (s32 i = 0; i < enemyCount; i++)
	{
		Enemy* e;

		e = enemies[i];

		if 
		(
			circle_collide(&_s->shadow, &e->shadow) &&
			e->isActive
		)
		{
			Node* returnSnake;
			Node* snake;

			snake = list_node_from_data_get(&e->snakesHit, &_s, sizeof(Snake*));
			returnSnake = list_node_from_data_get(&e->returnSnakesHit, &_s, sizeof(Snake*));

			if 
			(
				_s->isReturning &&
				!returnSnake

			)
			{
				list_append_new(&e->returnSnakesHit, &_s, sizeof(Snake*));
				enemy_damage(e, _s->damage, &_l->enemies, _l);
			}
			else if
			(
				!_s->isReturning &&
				!snake
			)
			{
				list_append_new(&e->snakesHit, &_s, sizeof(Snake*));
				enemy_damage(e, _s->damage, &_l->enemies, _l);
			}
		}
	}

	free(enemies);

	collectibles = (Collectible**)list_data_get(&_l->collectibles);
        collectibleCount = _l->collectibles.count;

        for (s32 i = 0; i < collectibleCount; i++)
        {
                Collectible* c;

                c = collectibles[i];

                if
                (
                        circle_collide(&_s->shadow, &c->shadow) &&
                        c->isActive
                )
                {
                        c->attachedSnake = _s;
                        c->isAttached = TRUE;
                }
        }

        free(collectibles);

	if (_s->isReturning)
	{
		f64 angle;
		
		angle = math_atan
		(
			_p->shadow.x,
			_p->shadow.y,
			_s->shadow.x,
			_s->shadow.y
		);

		_s->sprite.rotation = RAD_TO_DEG * (angle + PI);

		_s->returnTimer++;
		_s->returnTimer = CEIL(_s->returnTimer, PLAYER_SNAKE_RETURN_TIMER_MAX);

		_s->dX = cos(angle) * (PLAYER_SNAKE_VELOCITY * (_s->returnTimer * PLAYER_SNAKE_RETURN_VELOCITY_MULTIPLIER));
		_s->dY = sin(angle) * (PLAYER_SNAKE_VELOCITY * (_s->returnTimer * PLAYER_SNAKE_RETURN_VELOCITY_MULTIPLIER));

		if (circle_collide(&_p->shadow, &_s->shadow))
			player_snake_remove(_s, _p, _l);

	}
	else
	{
		_s->lifetime--;
		_s->redirectTimer--;

		if (_s->redirectTimer <= 0)
			_s->sprite.frame = SNAKE_NONE;

		if (_s->lifetime <= 0)
		{
			_s->dX = 0;
			_s->dY = 0;
			_s->isReturning = TRUE;
			_s->sprite.frame = SNAKE_RETURN;

			sound_play(_l->g->sounds[SOUND_SNAKE_RETURN], 7);
		}

	}
}

/* Draws a given snake. */
static void
player_snake_draw(Snake* _s, Player* _p, Play* _l)
{
	if (!_s->isReturning)
	{
		f64 angle;

		angle = _s->sprite.rotation;
		_s->arc.rotation = angle;

		_s->arc.x = _s->sprite.x - (cos(DEG_TO_RAD * angle) * PLAYER_SNAKE_ARC_OFFSET);
		_s->arc.y = _s->sprite.y - (sin(DEG_TO_RAD * angle) * PLAYER_SNAKE_ARC_OFFSET);

		sprite_draw(&_s->arc, _l->g->renderer);
	}

	shadow_draw(&_s->shadow, _l);
	sprite_draw(&_s->sprite, _l->g->renderer);
}

/* Adds a snake. */
void
player_snake_add(Player* _p, Play* _l)
{
	Snake snake;
	f64 angle;

	memset(&snake, '\0', sizeof(Snake));

	snake.sprite = snakeSprite;
	snake.arc = arcSprite;
	
	angle = math_atan
	(
		_p->sprite.x,
		_p->sprite.y,
		_l->g->mouse.x,
		_l->g->mouse.y
	);

	snake.shadow.r = PLAYER_SNAKE_SHADOW_RADIUS;
	snake.sprite.x = _p->sprite.x - (cos(angle) * PLAYER_SNAKE_OFFSET);
	snake.sprite.y = _p->sprite.y - (sin(angle) * PLAYER_SNAKE_OFFSET);

	snake.damage = PLAYER_SNAKE_DAMAGE_DEFAULT;
	snake.redirectCount = 0;
	snake.isReturning = FALSE;
	snake.returnTimer = 0;
	snake.animationTimer = PLAYER_SNAKE_ANIMATION_TIMER_MAX;
        snake.sprite.rotation = RAD_TO_DEG * angle;
        snake.dX = -cos(angle) * PLAYER_SNAKE_VELOCITY;
        snake.dY = -sin(angle) * PLAYER_SNAKE_VELOCITY;
        snake.lifetime = _p->lifetime;

	list_append_new(&_p->snakes, (void*)&snake, sizeof(Snake));
}

/* Resets the snake to a new angle/velocity (redirect). */
void
player_snake_reset(Snake* _s, Player* _p, Play* _l)
{
	f64 angle;

	angle = math_atan
	(
		_s->shadow.x,
		_s->shadow.y,
		_l->g->mouse.x,
		_l->g->mouse.y
	);
	
	_s->animationTimer = PLAYER_SNAKE_ANIMATION_TIMER_MAX;
	_s->sprite.rotation = RAD_TO_DEG * angle;
	_s->dX = -cos(angle) * (PLAYER_SNAKE_VELOCITY + (_s->redirectCount * PLAYER_SNAKE_REDIRECT_VELOCITY_MULTIPLIER));
	_s->dY = -sin(angle) * (PLAYER_SNAKE_VELOCITY + (_s->redirectCount * PLAYER_SNAKE_REDIRECT_VELOCITY_MULTIPLIER));
	
	_s->lifetime = (s32)(_p->lifetime * PLAYER_SNAKE_REDIRECT_LIFETIME_MULTIPLIER);
}

/* Initializes player snakes. */
void
player_snakes_init(Player* _p, Play* _l)
{
	Atlas atlas;

	atlas_init
	(
		&atlas,
		_l->g->textures[TEXTURE_PLAYER_SNAKE],
		PLAYER_SNAKE_ROWS,
		PLAYER_SNAKE_COLUMNS,
		PLAYER_SNAKE_FRAME_WIDTH,
		PLAYER_SNAKE_FRAME_HEIGHT
	);

	list_init(&_p->snakes);

	sprite_atlas_init
	(
		&snakeSprite,
		&atlas,
		ORIGIN_MIDDLE,
		PLAYER_SNAKE_WIDTH,
		PLAYER_SNAKE_HEIGHT
	);

	snakeSprite.frame = SNAKE_NONE;

	sprite_init
	(
		&arcSprite,
		_l->g->textures[TEXTURE_ARC],
		ORIGIN_MIDDLE,
		PLAYER_SNAKE_ARC_WIDTH,
		PLAYER_SNAKE_ARC_HEIGHT
	);
}

/* Clears all player snakes. */
void
player_snakes_clear(Player* _p, Play* _l)
{
	Snake** snakes;
	s32 snakeCount;

	snakes = (Snake**)list_data_get(&_p->snakes);
	snakeCount = _p->snakes.count;

	for (s32 i = 0; i < snakeCount; i++)
	{
		Snake* s;

		s = snakes[i];

		player_snake_remove(s, _p, _l);
	}
}

/* Frees snake list. */
void
player_snakes_free(Player* _p, Play* _l)
{
	list_free(&_p->snakes);
}

/* Updates all player snakes. */
void
player_snakes_update(Player* _p, Play* _l)
{
	Snake** snakes;

	snakes = (Snake**)list_data_get(&_p->snakes);

	for (s32 i = 0; i < _p->snakes.count; i++)
	{
		Snake* s;

		s = snakes[i];
	
		player_snake_update(s, _p, _l);
	}

	free(snakes);
}

/* Draws all player snakes. */
void
player_snakes_draw(Player* _p, Play* _l)
{
	Snake** snakes;

	snakes = (Snake**)list_data_get(&_p->snakes);

	for (s32 i = 0; i < _p->snakes.count; i++)
	{
		Snake* s;

		s = snakes[i];
	
		player_snake_draw(s, _p, _l);
	}

	free(snakes);
}

