/*
 * DESCRIPTION:
 * Enemy waves.
*/

#include "wave.h"

static void wave_set(Play* _p);
static void wave_next(Play* _p);
static void wave_clear(Play* _p);
static void wave_fail(Play* _p);
static void wave_reward(Play* _p);
static void wave_game_clear_begin(Play* _p);
static void wave_game_clear_begin(Play* _p);

/* Returns the cash reward you'll get at the end of a wave. */
static s32
wave_reward_get(Play* _p)
{
	f64 cash;

	cash = _p->player.waveEnemiesKilled * WAVE_REWARD_MULTIPLIER;

	if (_p->game > 1)
		cash *= ((_p->game - 1) * WAVE_REWARD_GAME_MULTIPLIER);

	return (s32)cash;
}

/* Sets the current wave of enemies. */
static void
wave_set(Play* _p)
{
	EnemyType type;
	s32 index;
	s32 enemyCountBase;
	s32 enemyCount;
	
	enemyCountBase = WAVE_ENEMY_COUNT_BASE;
	enemyCountBase *= (s32)(_p->wave * WAVE_ENEMY_COUNT_MULTIPLIER);
	enemyCountBase *= (s32)(_p->game * WAVE_ENEMY_COUNT_GAME_MULTIPLIER);

	enemyCount = random_s32_in_range_get
	(
		(s32)(enemyCountBase * WAVE_ENEMY_COUNT_BASE_MIN_MULTIPLIER),
		(s32)(enemyCountBase * WAVE_ENEMY_COUNT_BASE_MAX_MULTIPLIER)
	);

	list_clear(&_p->spawnEnemies);

	_p->player.waveEnemiesKilled = 0;
	_p->enemyMax = 0;

	for (s32 i = 0; i < enemyCount; i++)
	{
		s32 index;
		EnemyType type;

		switch (_p->wave)
		{
			case 1:
				index = random_s32_in_range_get(0, WAVE_ONE_ENEMY_TYPE_COUNT - 1);
				type = WAVE_ONE[index];
				break;
			case 2:
				index = random_s32_in_range_get(0, WAVE_TWO_ENEMY_TYPE_COUNT - 1);
				type = WAVE_TWO[index];
				break;
			case 3:
				index = random_s32_in_range_get(0, WAVE_THREE_ENEMY_TYPE_COUNT - 1);
				type = WAVE_THREE[index];
				break;
			case 4:
				index = random_s32_in_range_get(0, WAVE_FOUR_ENEMY_TYPE_COUNT - 1);
				type = WAVE_FOUR[index];
				break;
			case 5:
				index = random_s32_in_range_get(0, WAVE_FIVE_ENEMY_TYPE_COUNT - 1);
				type = WAVE_FIVE[index];
				break;
			case 6:
				index = random_s32_in_range_get(0, WAVE_SIX_ENEMY_TYPE_COUNT - 1);
				type = WAVE_SIX[index];
				break;
			case 7:
				index = random_s32_in_range_get(0, WAVE_SEVEN_ENEMY_TYPE_COUNT - 1);
				type = WAVE_SEVEN[index];
				break;
			case 8:
				index = random_s32_in_range_get(0, WAVE_EIGHT_ENEMY_TYPE_COUNT - 1);
				type = WAVE_EIGHT[index];
				break;
			case 9:
				index = random_s32_in_range_get(0, WAVE_NINE_ENEMY_TYPE_COUNT - 1);
				type = WAVE_NINE[index];
				break;
			case 10:
				index = random_s32_in_range_get(0, WAVE_TEN_ENEMY_TYPE_COUNT - 1);
				type = WAVE_TEN[index];
				break;
			default:
				break;
		}

		_p->enemyMax++;

		list_append_new(&_p->spawnEnemies, (void*)&type, sizeof(type));
	}

	_p->isBountyUpdate = TRUE;

	_p->spawnTimer = WAVE_GRACE_PERIOD_TIMER_MAX;
	_p->gracePeriodTimer = WAVE_GRACE_PERIOD_TIMER_MAX;
	_p->isGracePeriod = TRUE;
}

/* Entering between waves... */
static void
wave_between(Play* _p)
{
	wave_reward(_p);

	_p->player.waveEnemiesKilled = 0;
	_p->player.isSnakeUpdate = TRUE;

	_p->isBetweenWaves = TRUE;
	_p->isReady = FALSE;
	_p->player.isAbleToFire = FALSE;
	_p->isWaveUpdate = TRUE;
	_p->isUpgradesActive = TRUE;
	_p->isUpgradeUpdate = TRUE;
	_p->isBountyUpdate = TRUE;

	if (_p->player.isPowerUp)
		player_power_up_end(&_p->player, _p);
}

/* Grants a cash reward on wave end. */
static void
wave_reward(Play* _p)
{
	s32 cash;

	cash = wave_reward_get(_p);

	player_cash_acquire(&_p->player, cash, _p);

	sound_play(_p->g->sounds[SOUND_BULLIONS], -1);
}

/* Cleared the wave! */
static void
wave_clear(Play* _p)
{
	_p->isWaveCleared = TRUE;

	if (_p->wave + 1 > WAVE_MAX)
		wave_game_clear_begin(_p);
	else
	{
		message_wave_clear_add(_p->player.sprite.x, _p->player.sprite.y, _p);

		music_play(_p->g->music[MUSIC_WAVE_CLEAR], FALSE);

		_p->countdownTimer = WAVE_BETWEEN_TIMER_MAX;

		_p->isWaveUpdate = TRUE;
		_p->isBetweenWaves = TRUE;
	}
}

/* Failed the wave. */
static void
wave_fail(Play* _p)
{
	_p->isWaveFailed = TRUE;

	if (!_p->player.isBeingDefeated)
	{
		if (_p->wave + 1 > WAVE_MAX)
			wave_game_clear_begin(_p);
		else
		{
			message_wave_fail_add(_p->player.sprite.x, _p->player.sprite.y, _p);
			
			music_play(_p->g->music[MUSIC_WAVE_FAIL], FALSE);

			_p->countdownTimer = WAVE_BETWEEN_TIMER_MAX;

			_p->isWaveUpdate = TRUE;
			_p->isBetweenWaves = TRUE;
		
		}

	}
}

/* Goes to the next wave. */
static void
wave_next(Play* _p)
{
	_p->isUpgradesActive = FALSE;

	_p->player.isAbleToFire = TRUE;

	_p->isBetweenWaves = FALSE;

	_p->wave++;

	_p->spawnTimer = WAVE_GRACE_PERIOD_TIMER_MAX;

	_p->isWaveCleared = FALSE;
	_p->isWaveFailed = FALSE;
	_p->isWaveUpdate = TRUE;

	wave_set(_p);
}

/* You just beat the game! */
static void
wave_game_clear_begin(Play* _p)
{
	_p->isBetweenWaves = TRUE;
	_p->isWaveFailed = FALSE;
	_p->isWaveCleared = FALSE;
	_p->isGameClearing = TRUE;
	_p->gameClearTimer = WAVE_GAME_CLEAR_TIMER_MAX;

	list_clear(&_p->enemies);

	player_snakes_clear(&_p->player, _p);

	message_game_clear_add(_p->player.sprite.x, _p->player.sprite.y, _p);

	music_play(_p->g->music[MUSIC_WIND], TRUE);
}

/* You beat the game! */
static void
wave_game_clear(Play* _p)
{
	_p->isGameCleared = TRUE;
}

/* Initializes waves. */
void
wave_init(Play* _p)
{
	list_init(&_p->spawnEnemies);

	_p->wave = WAVE_DEFAULT;
	_p->game = GAME_DEFAULT;
	_p->isWaveCleared = FALSE;
	_p->isGameCleared = FALSE;

	wave_set(_p);
}

/* New Game+! */
void
wave_new_game_plus(Play* _p)
{
	list_clear(&_p->collectibles);

	_p->game++;
	_p->wave = WAVE_DEFAULT;
	_p->isWaveCleared = FALSE;
	_p->isGameCleared = FALSE;
	_p->isGameClearing = FALSE;
	_p->isWaveFailed = FALSE;
	_p->isWaveUpdate = TRUE;
	_p->isReady = FALSE;
	_p->isBetweenWaves = FALSE;
	_p->isUpgradesActive = FALSE;

	_p->player.sprite.x = PLAYER_SPRITE_X_DEFAULT;
	_p->player.sprite.y = PLAYER_SPRITE_Y_DEFAULT;

	_p->player.isAbleToFire = TRUE;
	_p->player.isAbleToMove = TRUE;

	_p->player.snakeReserve = _p->player.snakeMax;
	_p->player.redirects = _p->player.redirectMax;

	music_play(_p->g->music[MUSIC_PLAY], TRUE);

	wave_set(_p);
}


/* Updates waves. */
void
wave_update(Play* _p)
{
	if (!_p->isGameClearing)
	{
		if (_p->isGracePeriod)
		{
			_p->gracePeriodTimer--;

			if (_p->gracePeriodTimer <= 0)
				_p->isGracePeriod = FALSE;
		}

		if (_p->isWaveUpdate)
			_p->isWaveUpdate = FALSE;

		if (_p->isBetweenWaves)
		{
			if (!_p->isWaveClearMusicComplete)
			{
				if (!music_playing())
				{
					_p->isWaveClearMusicComplete = TRUE;
					music_play(_p->g->music[MUSIC_WIND], TRUE);
					wave_between(_p);
				}
			}
			else
			{
				if (_p->isReady)
				{
					_p->isWaveClearMusicComplete = FALSE;

					music_play(_p->g->music[MUSIC_PLAY], TRUE);
					
					_p->isReady = FALSE;
					_p->isBetweenWaves = FALSE;

					wave_next(_p);
				}
			}
		}
		else
		{
			if 
			(
				_p->enemies.count <= 0 && 
				_p->spawnEnemies.count <= 0
			)
				_p->isWaveCleared = TRUE;

			if (_p->isWaveFailed)
				wave_fail(_p);
			else if (_p->isWaveCleared)
				wave_clear(_p);
		}
	}
	else
	{
		_p->gameClearTimer--;

		_p->player.isAbleToMove = FALSE;
		_p->player.isAbleToFire = FALSE;

		if (_p->gameClearTimer <= 0)
			wave_game_clear(_p);
	}
}
